/*
CS255 HW5 Spring 2018
SJSU STUDENTS

*/

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.io.LineNumberReader;
import java.io.Reader;
import java.util.Arrays;
import java.util.HashSet;
import java.util.Set;


public class ApproxSetCover {

	static boolean debug = false; // enable this for debugging

	HashSet<Integer> X; // first line of input file
	HashSet<?>[] F;	// all Sets except F, from the input file
	HashSet<?>[] C; // Set cover found
	static int length_of_bin_string = 0;
	static int size_of_F = 0;

	public static void main(String arg[]) {

		String usage = "Usage: java ApproxSetCover <inputfile>" +
				"\ninputfile = mandatory arg for input sets";
		if(arg.length != 1) {
				System.out.println(usage);
				return;
		}

		ApproxSetCover as = new ApproxSetCover();
		if(!as.initSets(arg[0])) {
			System.out.println("Failed to parse input file" + arg[0]);
			return;
		}

		if(debug) as.printInputSets();

		/********************************
		 * Greedy set cover
		 ********************************/
		HashSet<Integer>[] C = as.GREEDY_SET_COVER();
		String bin[] = convertSetArrayToBinStringArray(C);
		if(bin != null) {
			//System.out.println("GREEDY_SET_COVER returned ");
			for(String s: bin) {
				if(s!=null)
					System.out.println(s);
			}
		} else {
			System.out.println("GREEDY_SET_COVER returned Null. Set Cover not possible.");
		}

		/********************************
		 * brute-force set cover
		 ********************************/
		System.out.println("");
		C = as.BRUTE_FORCE_SET_COVER();
		bin = convertSetArrayToBinStringArray(C);
		if(bin != null) {
			//System.out.println("BRUTE_FORCE_SET_COVER returned ");
			for(String s: bin) {
				if(s!=null)
					System.out.println(s);
			}
		} else {
			System.out.println("BRUTE_FORCE_SET_COVER returned Null. Set Cover not possible.");
		}

	}

	/********************************
	 * Greedy set cover
	 ********************************/

	private HashSet<Integer>[] GREEDY_SET_COVER() {
		/*GREEDY-SET-COVER(X, F)
		1 U := X
		2 C := null
		3 while U != 0
		4     select an S member of F that maximizes {S intersection U}
		5     U := U - S
		6     C := C Union {S}
		7 return C
		*/

		HashSet<Integer> U = new HashSet<Integer>(X);
		//System.out.println("U " + U);
		C = new HashSet<?>[size_of_F];
		int c_index = 0;
		HashSet<Integer> max_S = null;
		while(U.size() != 0) {
			max_S = MaxSet(U,F);
			U = diff(U,max_S);
			C = union(C, c_index, max_S);
			if(C == null)
				return null;
			c_index++;
		}
		return (HashSet<Integer>[]) C;
	}

	private HashSet<Integer> MaxSet(HashSet<Integer> U, HashSet<?>[] F) {
		/*
		 * line 4 of the GREEDY-SET-COVER
		 * select an S member of F that maximizes {S intersection U}
		 *
		 * Loop over all sets of F and return a set which has maximam size of {S intersection U}
		 *
		 */
		HashSet<Integer> max = null ,temp;

		for(HashSet<?> s: F) {
			//System.out.println("max F = " + s);
			if( s != null) {
				temp = interSection((HashSet<Integer>) s,U);
				if( max == null) {
					max = temp;
				} else {
					if(temp.size() > max.size())
						max = (HashSet<Integer>) s;
				}
			}
		}
		//System.out.println("max returning = " + max);
		return max;
	}

	/* generate an array of sets from f, with indexes given by array comb */
	private HashSet<Integer>[] genProbableSetCoverFromComb(HashSet<?>[] f, int[] comb) {
		HashSet<?>[] temp = new HashSet<?>[comb.length];
		int i = 0;
		for(int n: comb) {
			temp[i] = f[n];
			i++;
		}
		return (HashSet<Integer>[]) temp;
	}

	/* check if array of Sets can be cover of the "cover" passed.
	 * return true if Sets can be cover else return false */
	private boolean checkCover(HashSet<Integer> cover, HashSet<Integer>[] Sets) {
		if(cover == null) {
			if(debug) System.out.println("cover cannot be null in checkCover");
			return false;
		}
		if ((Sets == null) || (Sets.length==0)) {
			if(debug) System.out.println("Sets cannot be null in checkCover");
			return false;
		}

		if(debug) System.out.println("	Checking for cover " + cover + " under sets ");

		HashSet<Integer> remaining = cover;
		for(HashSet<Integer> s: Sets) {
			if(s !=null) {
				if(debug) System.out.println("	s = " + s);
				remaining = diff(remaining,s);
			}
		}

		//System.out.println("	remaining " + remaining);

		if (remaining.size() == 0) {
			if(debug) System.out.println("	Checking for cover " + cover + " passed ");
			return true;
		}

		if(debug) System.out.println("	Checking for cover " + cover + " failed ");
		return false;
	}

	/* adds set max_S to array of set c */
	private HashSet<?>[] union(HashSet<?>[] c, int c_index, HashSet<Integer> max_S) {
		//System.out.println("c len= " + c.length + " and c_index " + c_index );
		if(c_index >= c.length )
			return null;
		c[c_index] = max_S;
		return c;
	}

	/* returns set a - b */
	private HashSet<Integer> diff(HashSet<Integer> a, HashSet<Integer> b) {
		if((a == null) || (b==null))
			return null;
		HashSet<Integer> c =  new HashSet<Integer>(a);
		c.removeAll(b);
		//System.out.println("diff of a = " + a + " and b " + b + " is " + c );
		return c;
	}

	/* returns set a intersection set b */
	private HashSet<Integer> interSection(HashSet<Integer> a, HashSet<Integer> b) {
		HashSet<Integer> c =  new HashSet<Integer>(a);
		c.retainAll(b);
		//System.out.println("interSection of a = " + a + " and b " + c + " is " + c );
		return c;
	}

	/* take array of sets, each of which has decimal numbers and output and array of
	 * strings which has binary encoding of the sets.
	 */
	private static String[] convertSetArrayToBinStringArray(HashSet<Integer>[] setArray) {
		if(setArray == null)
			return null;
		String str[] = new String[setArray.length];
		int index = 0;
		for(HashSet<Integer> s: setArray) {
			if(s!= null) {
				str[index] = convertSetToBinString(s);
				//System.out.println("input set = " + s + " output str = " + str[index]);
				index++;
			}
		}
		return str;
	}

	/* replace char c at position pos in String s */
	public static String replaceCharAt(String s, Integer pos, char c) {
		   return s.substring(0,pos) + c + s.substring(pos+1);
	}

	/* conver single set to binary encoding */
	private static String convertSetToBinString(HashSet<Integer> s) {
		String temp = "";
		for(int i = 0; i < length_of_bin_string ;i++) {
			temp += "0";
		}
		if(s != null) {
			for (Integer j : s) {
	            temp = replaceCharAt(temp, j-1,'1');
	        }
		}
		return temp;
	}

	/* print input sets for debugging */
	private void printInputSets() {
		System.out.println("X = " + X);
		for(HashSet<?> s: F) {
			System.out.println("F = " + s);
		}
	}

	/*return true if string st is valid i.e. not null, not empty and contains only 0s or 1s
	 * else return false as its invalid string */
	private boolean isValid(String st) {
		//System.out.println("isValid st = " + st );
		if ((st == null) || (st.length() == 0) || st.matches("\\s+"))
			return false;
		return st.matches("[01]+");
	}

	/* read the file fileName and initialize the X and F sets.
	 * read each line and if its valid then convert the line to decimal encoded sets
	 * and init X and F with the Sets.
	 *
	 * return true if sets get generated correctly, else false.
	 *
	 * */
	private boolean initSets(String fileName) {
		/* fileName is input file */

		BufferedReader br = null;
		Reader r = null;
		String st;
		boolean result = false;
		try {
			/* read the file line by line */
			File file = new File(fileName);
			r = new FileReader(file);
			br = new BufferedReader(r);
			LineNumberReader lineNumberReader = new LineNumberReader(new FileReader(file));
			lineNumberReader.skip(Long.MAX_VALUE);
			//System.out.println("num of lines in file " + lineNumberReader.getLineNumber()); // reads one
			size_of_F = lineNumberReader.getLineNumber();
			st = br.readLine();
			while (!isValid(st)) {
				System.out.println("skipping " + st);
				st = br.readLine();
			}
			/* store len to convert back set to bin string */
			length_of_bin_string = st.length();
			X = (HashSet<Integer>) converToSet(st);
			if (X == null)
				return false;

			HashSet<?>[] tempF = new HashSet<?>[size_of_F];
			HashSet<Integer> temp;
			int i = 0;
			st = br.readLine();
			while ( st != null) {
				if (isValid(st)) {
					temp = (HashSet<Integer>) converToSet(st);
					if (temp == null)
						return false;
					else {
						tempF[i] = (HashSet<Integer>) temp;
						i++;
					}
				}
				st = br.readLine();
			}

			/* Create F from tempF by removing null Set for invalid or empty lines. */
			int count = 0;
			F = new HashSet<?>[i];
			for(HashSet<?> s: tempF) {
				if( (s != null) && (s.size()>0) ) {
					F[count] = s;
					count++;
				}
			}
			result = true;
			size_of_F = i; // number of valid lines and so its number of valid F sets
		} catch (Exception e) {
			System.out.println(e);
			result =  false;
		}
		try {
			br.close();
		} catch (IOException e) {
			e.printStackTrace();
		}

		return result;
	}

	/* re-encode set to binary string to print output */
	private Set<Integer> converToSet(String st) {
		 Set<Integer> temp = new HashSet<Integer>();
		 int i = 1;
		 for (char ch: st.toCharArray()) {
			 if(ch == '1')
				 temp.add(i);
			  i++;
		 }
		//System.out.println("input st = " + st + " output set = " + temp);
		return temp;
	}

	class CombGen {
		private int n, r;
		private int[] source;
		private boolean isRemaining;

		public CombGen(int[] source, int n, int r) {
			this.n = n;
			this.r = r;
			this.source = source;
			this.isRemaining = true;
		}

		private int bar() {
			for (int index = r - 1; index >= 0; index--)
				if (source[index] < n - r + index)
					return index;
			return -1;
		}

		public boolean getRemaining() {
			return isRemaining;
		}

		private void moveSource() {
			int index = bar();
			if (index >= 0) {
				source[index] = source[index] + 1;
				for (int j = index + 1; j < r; j++)
					source[j] = source[j - 1] + 1;
			} else {
				isRemaining = false;
			}
		}

		public int[] next() {
			if (!isRemaining) {
				return null;
			}
			int[] comb = new int[r];
			for (int i = 0; i < r; i++) {
				comb[i] = source[i];
			}
			moveSource();
			return comb;
		}
	}

	/********************************
	 * brute-force set cover
	 ********************************/
	private HashSet<Integer>[] BRUTE_FORCE_SET_COVER() {

		/*
		 * start with single set and then 2 sets and so on unto size_of_F and see if the
		 * # of sets cover the original set
		 */


		int min = size_of_F;
		HashSet<Integer>[] min_Cover = null;
		for (int r = 1; r <= size_of_F; r++) {
			/* generate combinations of size r from arr */
			if(debug) System.out.println("Generate combinations of length " + r);
			int source[] = new int[size_of_F];
		    for (int temp_i = 0; temp_i < size_of_F; temp_i++)
		    	source[temp_i] = temp_i;
			CombGen c = new CombGen(source, size_of_F, r);
			while (c.getRemaining()) {
				int[] next = c.next();
				if(debug) System.out.println(Arrays.toString(next));
				/*
				 * create an array of sets bases on comb and see if that array of sets is a set
				 * cover of X.
				 */
				HashSet<Integer>[] input = genProbableSetCoverFromComb(F, next);
				if (checkCover(X, input)) {
					if(debug) {
						System.out.println("========= One possible set cover found =========");
						for (HashSet<Integer> s : input) {
							System.out.println("F = " + s);
						}
						System.out.println("=============================================");
					}
					if (input.length < min) {
						min_Cover = input;
						min = input.length;
						if(debug) {
							System.out.println("========= So far optimal set cover: =========");
							for (HashSet<Integer> s : input) {
								System.out.println("F = " + s);
							}
							System.out.println("=============================================");
						}
					}
				}
			}
		}
		return min_Cover;
	}
}
